from .execution_engine import ExecutionEngine
from ..worker import worker_process


# pylint: disable=too-few-public-methods
class LocalExecutionEngine(ExecutionEngine):
    def __call__(self, timeout, pending_work, config):
        for work_record in pending_work:
            yield worker_process(work_record, timeout, config)



"""Functionality related to Python's import mechanisms."""

import contextlib
import sys
from importlib.abc import MetaPathFinder
from importlib.machinery import ModuleSpec


class ASTLoader:  # pylint:disable=old-style-class,too-few-public-methods

    """
    An `importlib.abc.Loader` which loads an AST for a particular name.

    You construct this with an AST and a module name. The
    `exec_module` method simply compiles the AST with the name and
    execs the resulting code against the provided module dict.

    In practice, this is how cosmic-ray loads mutated ASTs for modules.
    """

    def __init__(self, ast, name):
        self._ast = ast
        self._name = name

    def create_module(self,  # pylint: disable=no-self-use
                      spec):  # pylint: disable=unused-argument
        return None

    def exec_module(self, mod):
        compiled = compile(self._ast, self._name, 'exec')
        exec(compiled, mod.__dict__)  # pylint:disable=exec-used


class ASTFinder(MetaPathFinder):  # pylint:disable=too-few-public-methods

    """
    An `importlib.ast.MetaPathFinder` that associates a module name
    with an AST.

    Construct this by passing a module name and associated AST. When
    the finder is asked for that name, it will return a loader that
    produces the code equivalent of the AST.

    We use this to inject mutated ASTs into tests.
    """

    def __init__(self, fullname, ast):
        self._fullname = fullname
        self._ast = ast

    def find_spec(self, fullname,
                  path, target=None):  # pylint:disable=unused-argument
        if fullname == self._fullname:
            return ModuleSpec(fullname,
                              ASTLoader(self._ast, fullname))


@contextlib.contextmanager
def preserve_modules():
    """Remember the state of sys.modules on enter and reset it on exit.
    """
    original_mods = dict(sys.modules)
    try:
        yield
    finally:
        del_mods = {m for m in sys.modules if m not in original_mods}
        for mod in del_mods:
            del sys.modules[mod]


@contextlib.contextmanager
def using_ast(module_name, module_ast):
    """Create a new Finder as a context-manager.

    This creates a new Finder which loads the AST `module_ast` when
    `module_name` is requested. It installs this finder at the head of
    `sys.meta_path` for the duration of the with-block, yielding the Finder in
    the with-statement. After the with-block, the Finder is uninstalled.

    Note that this does *not* attempt to adjust `sys.modules` in any way. You
    should make sure to clear out any existing references to `module_name`
    before running this (e.g. with `preserve_modules()` or something similar).

    """
    finder = ASTFinder(module_name, module_ast)
    sys.meta_path = [finder] + sys.meta_path
    try:
        yield finder
    finally:
        sys.meta_path.remove(finder)



import ast


class Operator(ast.NodeTransformer):
    """
    A base class for all mutation operators.

    This takes care of the basic book-keeping that all operators need
    to do. All operators *must* derive from this class since that's
    how we keep track of them.
    """

    def __init__(self, core):
        self._core = core

    @property
    def core(self):
        """The core behavior of the operator."""
        return self._core

    def repr_args(self):
        return self.core.repr_args()

    def visit_mutation_site(self, node, num_mutations=1):
        """Subclasses call this when they encounter a node they can
         potentially mutate.

        This functions delegates to the core, letting it do whatever it needs
        to do.
        """
        return self.core.visit_mutation_site(node, self, num_mutations)

    def mutate(self, node, idx):
        """Mutate a node in an operator-specific manner.

        Return the new, mutated node. Return `None` if the node has
        been deleted. Return `node` if there is no mutation at all for
        some reason.
        """
        raise NotImplementedError(
            'Mutation operators must implement "mutate()".')

    def __repr__(self):
        repr_args = [('core', self.core.__class__.__name__)]
        repr_args.extend(self.core.repr_args())
        args = ['{}={}'.format(k, v)
                for k, v
                in repr_args]

        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(args))



import ast
import sys

from .operator import Operator
from ..util import build_mutations

OPERATORS = (ast.Add, ast.Sub, ast.Mult, ast.Div, ast.FloorDiv, ast.Mod,
             ast.Pow, ast.LShift, ast.RShift, ast.BitOr, ast.BitXor,
             ast.BitAnd)

# todo: this often leads to unsupported syntax
if sys.version_info >= (3, 5):
    OPERATORS = OPERATORS + (ast.MatMult,)


def _to_ops(from_op):  # pylint: disable=unused-argument
    """The sequence of operators which `from_op` could be mutated to."""
    for to_op in OPERATORS:
        yield to_op


class MutateBinaryOperator(Operator):
    """An operator that modifies binary operators."""

    def visit_BinOp(self, node):  # pylint: disable=invalid-name
        """
            http://greentreesnakes.readthedocs.io/en/latest/nodes.html#BinOp
        """
        return self.visit_mutation_site(
            node,
            len(build_mutations([node.op], _to_ops)))

    def mutate(self, node, idx):
        _, to_op = build_mutations([node.op], _to_ops)[idx]
        node.op = to_op()
        return node



"""Functions for calculate certain kinds of reports.
"""

from cosmic_ray.testing.test_runner import TestOutcome
from cosmic_ray.worker import WorkerOutcome


def _print_item(work_record, full_report):
    data = work_record.data
    outcome = work_record.worker_outcome
    if outcome in [WorkerOutcome.NORMAL, WorkerOutcome.EXCEPTION]:
        outcome = work_record.test_outcome
    ret_val = [
        'job ID {}:{}:{}'.format(
            work_record.job_id,
            outcome,
            work_record.module),
        'command: {}'.format(work_record.command_line or '')
    ]
    if outcome == TestOutcome.KILLED and not full_report:
        ret_val = []
    elif work_record.worker_outcome == WorkerOutcome.TIMEOUT:
        if full_report:
            ret_val.append("timeout: {:.3f} sec".format(data))
        else:
            ret_val = []
    elif work_record.worker_outcome in [WorkerOutcome.NORMAL,
                                        WorkerOutcome.EXCEPTION]:
        ret_val += data
        ret_val += work_record.diff

    # for presentation purposes only
    if ret_val:
        ret_val.append('')

    return ret_val


def is_killed(record):
    """Determines if a WorkRecord should be considered "killed".
    """
    if record.worker_outcome == WorkerOutcome.TIMEOUT:
        return True
    elif record.worker_outcome == WorkerOutcome.NORMAL:
        if record.test_outcome == TestOutcome.KILLED:
            return True
    return False


def create_report(records, show_pending, full_report=False):
    total_jobs = 0
    pending_jobs = 0
    kills = 0
    for item in records:
        total_jobs += 1
        if item.worker_outcome is None:
            pending_jobs += 1
        if is_killed(item):
            kills += 1
        if (item.worker_outcome is not None) or show_pending:
            yield from _print_item(item, full_report)

    completed_jobs = total_jobs - pending_jobs

    yield 'total jobs: {}'.format(total_jobs)

    if completed_jobs > 0:
        yield 'complete: {} ({:.2f}%)'.format(
            completed_jobs, completed_jobs / total_jobs * 100)
        yield 'survival rate: {:.2f}%'.format(
            (1 - kills / completed_jobs) * 100)
    else:
        yield 'no jobs completed'


def survival_rate(records):
    """Calcuate the survival rate for a sequence of WorkRecords.
    """
    total_jobs = 0
    pending_jobs = 0
    kills = 0
    for item in records:
        total_jobs += 1
        if item.worker_outcome is None:
            pending_jobs += 1
        if is_killed(item):
            kills += 1

    completed_jobs = total_jobs - pending_jobs

    if not completed_jobs:
        rate = 0
    else:
        rate = (1 - kills / completed_jobs) * 100

    return rate



import ast

from .operator import Operator


class ReplaceBreakWithContinue(Operator):
    def visit_Break(self, node):  # pylint: disable=invalid-name
        return self.visit_mutation_site(node)

    def mutate(self, node, _):
        """Replace a Break node with a Continue node."""
        assert isinstance(node, ast.Break)
        return ast.Continue()


class ReplaceContinueWithBreak(Operator):
    def visit_Continue(self, node):  # pylint: disable=invalid-name
        return self.visit_mutation_site(node)

    def mutate(self, node, _):
        """Replace a Continue node with a Break node."""
        assert isinstance(node, ast.Continue)
        return ast.Break()



import unittest
from itertools import chain

from .test_runner import TestRunner


class UnittestRunner(TestRunner):  # pylint: disable=too-few-public-methods
    """A TestRunner using `unittest`'s discovery mechanisms.

    This treats the `test_args` as a directory name. This discovers
    all tests under that directory and executes them.

    All elements in `test_args` after the first are ignored.

    """

    def __init__(self, test_args):  # pylint: disable=useless-super-delegation
        super().__init__(test_args)

    def _run(self):
        suite = unittest.TestLoader().discover(self.test_args)
        result = unittest.TestResult()
        result.failfast = True
        suite.run(result)

        return (
            result.wasSuccessful(),
            [r[1] for r in chain(result.errors, result.failures)])



import json
import sys
import xml.etree.ElementTree

import docopt

from cosmic_ray.reporting import create_report, is_killed, survival_rate
from cosmic_ray.testing.test_runner import TestOutcome
from cosmic_ray.work_record import WorkRecord
from cosmic_ray.worker import WorkerOutcome


def format_survival_rate():
    """cr-rate

Usage: cr-rate

Read JSON work-records from stdin and print the survival rate.
"""
    records = (WorkRecord(json.loads(line)) for line in sys.stdin)
    print('{:.2f}'.format(survival_rate(records)))


def report():
    """cr-report

Usage: cr-report [--full-report] [--show-pending]

Print a nicely formatted report of test results and some basic statistics.

options:
    --full-report   Show test output and mutation diff for killed mutants
    --show-pending  Display results for incomplete tasks
"""

    arguments = docopt.docopt(report.__doc__, version='cr-format 0.1')
    full_report = arguments['--full-report']
    show_pending = arguments['--show-pending']
    records = (WorkRecord(json.loads(line)) for line in sys.stdin)
    for line in create_report(records, show_pending, full_report):
        print(line)


def _create_element_from_item(work_record):
    data = work_record.data
    sub_elem = xml.etree.ElementTree.Element('testcase')

    sub_elem.set('classname', work_record.job_id)
    sub_elem.set('line', str(work_record.line_number))
    sub_elem.set('file', work_record.module)
    if work_record.command_line:
        sub_elem.set('name', str(work_record.command_line))

    outcome = work_record.worker_outcome

    if outcome == WorkerOutcome.TIMEOUT:
        error_elem = xml.etree.ElementTree.SubElement(sub_elem, 'error')
        error_elem.set('message', "Timeout: {:.3f} sec".format(data))
    elif outcome == WorkerOutcome.EXCEPTION:
        error_elem = xml.etree.ElementTree.SubElement(sub_elem, 'error')
        error_elem.set('message', "Worker has encountered exception")
        error_elem.text = str(data) + "\n".join(work_record.diff)
    elif _evaluation_success(outcome, work_record):
        failure_elem = xml.etree.ElementTree.SubElement(sub_elem, 'failure')
        failure_elem.set('message', "Mutant has survived your unit tests")
        failure_elem.text = str(data) + "\n".join(work_record.diff)

    return sub_elem


def _evaluation_success(outcome, work_record):
    return outcome == WorkerOutcome.NORMAL and \
           work_record.test_outcome in [TestOutcome.SURVIVED,
                                        TestOutcome.INCOMPETENT]


def _create_xml_report(records):
    total_jobs = 0
    errors = 0
    failed = 0
    root_elem = xml.etree.ElementTree.Element('testsuite')
    for item in records:
        total_jobs += 1
        if item.worker_outcome is None:
            errors += 1
        if is_killed(item):
            failed += 1
        if item.worker_outcome is not None:
            subelement = _create_element_from_item(item)
            root_elem.append(subelement)

    root_elem.set('errors', str(errors))
    root_elem.set('failures', str(failed))
    root_elem.set('skips', str(0))
    root_elem.set('tests', str(total_jobs))
    return xml.etree.ElementTree.ElementTree(root_elem)


def report_xml():
    """cr-xml

Usage: cr-xml

Print an XML formatted report of test results for continuos integration systems
"""
    records = (WorkRecord(json.loads(line)) for line in sys.stdin)
    xml_elem = _create_xml_report(records)
    xml_elem.write(sys.stdout.buffer, encoding='utf-8', xml_declaration=True)



"""Configuration module."""
import sys

import yaml


def load_config(filename=None):
    """Load a configuration from a file or stdin.

    If `filename` is `None`, then this reads its configuration from stdin.

    Returns: A configuration dict.
    """
    if filename is None:
        return yaml.safe_load(sys.stdin)

    with open(filename, mode='rt') as handle:
        return yaml.safe_load(handle)


def serialize_config(config):
    """Convert a configuration dict into a string.

    This is complementary with `load_config`.
    """
    return yaml.dump(config)


def get_db_name(session_name):
    """Determines the filename for a session.

    Basically, if `session_name` ends in ".json" this return `session_name`
    unchanged. Otherwise it return `session_name` with ".json" added to the
    end.
    """
    if session_name.endswith('.json'):
        return session_name
    return '{}.json'.format(session_name)


class ConfigError(Exception):
    """An error occurred while creating configuration."""
    pass



import ast

from .operator import Operator


class NumberReplacer(Operator):
    """An operator that modifies numeric constants."""

    def visit_Num(self, node):  # noqa # pylint: disable=invalid-name
        return self.visit_mutation_site(node)

    def mutate(self, node, _):
        """Modify the numeric value on `node`."""
        new_node = ast.Num(n=node.n + 1)
        return new_node



"""This is the command-line program for cosmic ray.

Here we manage command-line parsing and launching of the internal
machinery that does mutation testing.
"""
import itertools
import json
import logging
import os
import pprint
import signal
import subprocess
import sys

import docopt
import docopt_subcommands as dsc

import cosmic_ray.commands
import cosmic_ray.counting
import cosmic_ray.modules
import cosmic_ray.plugins
import cosmic_ray.worker
from cosmic_ray.config import ConfigError, get_db_name, load_config
from cosmic_ray.testing.test_runner import TestOutcome
from cosmic_ray.timing import Timer
from cosmic_ray.util import redirect_stdout
from cosmic_ray.work_db import use_db, WorkDB

LOG = logging.getLogger()


@dsc.command()
def handle_baseline(args):
    """usage: cosmic-ray baseline <config-file>

Run an un-mutated baseline of the specific configuration. This is largely like
running a "worker" process, with the difference that a baseline run doesn't
mutate the code.

    """
    sys.path.insert(0, '')

    config = load_config(args['<config-file>'])

    test_runner = cosmic_ray.plugins.get_test_runner(
        config['test-runner']['name'],
        config['test-runner']['args'])

    work_record = test_runner()
    # note: test_runner() results are meant to represent
    # status codes when executed against mutants.
    # SURVIVED means that the test suite executed without any error
    # hence CR thinks the mutant survived. However when running the
    # baseline execution we don't have mutations and really want the
    # test suite to report PASS, hence the comparison below!
    if work_record.test_outcome != TestOutcome.SURVIVED:
        # baseline failed, print whatever was returned
        # from the test runner and exit
        LOG.error('baseline failed')
        print(''.join(work_record.data))
        sys.exit(2)


@dsc.command()
def handle_new_config(args):
    """usage: cosmic-ray new-config <config-file>

Create a new config file.
    """
    config = cosmic_ray.commands.new_config()
    with open(args['<config-file>'], mode='wt') as handle:
        handle.write(config)


@dsc.command()
def handle_init(args):
    """usage: cosmic-ray init <config-file> <session-file>

Initialize a mutation testing session from a configuration. This primarily
creates a session - a database of "work to be done" - which describes all of
the mutations and test runs that need to be executed for a full mutation
testing run. The configuration specifies the top-level module to mutate, the
tests to run, and how to run them.

This command doesn't actually run any tests. Instead, it scans the
modules-under-test and simply generates the work order which can be executed
with other commands.

The `session-file` is the filename for the database in which the work order
will be stored.
    """
    # This lets us import modules from the current directory. Should probably
    # be optional, and needs to also be applied to workers!
    sys.path.insert(0, '')

    config_file = args['<config-file>']

    if not os.path.exists(config_file):
        handle_new_config(args)

    config = load_config(config_file)

    if 'timeout' in config:
        timeout = float(config['timeout'])
    elif 'baseline' in config:
        baseline_mult = float(config['baseline'])
        # TODO: Should not be assertion
        assert baseline_mult is not None
        command = 'cosmic-ray baseline {}'.format(
            args['<config-file>'])

        # We run the baseline in a subprocess to more closely emulate the
        # runtime of a worker subprocess.
        with Timer() as timer:
            subprocess.check_call(command.split())

        timeout = baseline_mult * timer.elapsed.total_seconds()
    else:
        raise ConfigError(
            "Config must specify either baseline or timeout")

    LOG.info('timeout = %f seconds', timeout)

    modules = set(
        cosmic_ray.modules.find_modules(
            cosmic_ray.modules.fixup_module_name(config['module']),
            config.get('exclude-modules', None)))

    LOG.info('Modules discovered: %s', [m.__name__ for m in modules])

    db_name = get_db_name(args['<session-file>'])

    with use_db(db_name) as database:
        cosmic_ray.commands.init(
            modules,
            database,
            config,
            timeout)


@dsc.command()
def handle_config(args):
    """usage: cosmic-ray config <session-file>

Show the configuration for in a session.
    """
    session_file = get_db_name(args['<session-file>'])
    with use_db(session_file) as database:
        config, _ = database.get_config()
        print(json.dumps(config))


@dsc.command()
def handle_exec(args):
    """usage: cosmic-ray exec <session-file>

Perform the remaining work to be done in the specified session. This requires
that the rest of your mutation testing infrastructure (e.g. worker processes)
are already running.
    """

    session_file = get_db_name(
        args.get('<session-file>'))
    cosmic_ray.commands.execute(session_file)


@dsc.command()
def handle_dump(args):
    """usage: cosmic-ray dump <session-file>

JSON dump of session data. This output is typically run through other programs
to produce reports.
    """
    session_file = get_db_name(args['<session-file>'])

    with use_db(session_file, WorkDB.Mode.open) as database:
        for record in database.work_records:
            print(json.dumps(record))


@dsc.command()
def handle_counts(args):
    """usage: {program} counts <config-file>

Count the number of tests that would be run for a given testing configuration.
This is mostly useful for estimating run times and keeping track of testing
statistics.
"""
    config = load_config(args['<config-file>'])

    sys.path.insert(0, '')

    modules = cosmic_ray.modules.find_modules(
        cosmic_ray.modules.fixup_module_name(config['module']),
        config.get('exclude-modules', []))

    operators = cosmic_ray.plugins.operator_names()

    counts = cosmic_ray.counting.count_mutants(modules, operators)

    print('[Counts]')
    pprint.pprint(counts)
    print('\n[Total test runs]\n',
          sum(itertools.chain(
              *(d.values() for d in counts.values()))))


@dsc.command()
def handle_test_runners(config):  # pylint: disable=unused-argument
    """usage: {program} test-runners

List the available test-runner plugins.
"""
    print('\n'.join(cosmic_ray.plugins.test_runner_names()))
    return 0


@dsc.command()
def handle_operators(config):  # pylint: disable=unused-argument
    """usage: {program} operators

List the available operator plugins.
"""
    print('\n'.join(cosmic_ray.plugins.operator_names()))
    return 0


@dsc.command()
def handle_worker(args):
    """usage: {program} worker \
    [options] <module> <operator> <occurrence> [<config-file>]

Run a worker process which performs a single mutation and test run. Each worker
does a minimal, isolated chunk of work: it mutates the <occurence>-th instance
of <operator> in <module>, runs the test suite defined in the configuration,
prints the results, and exits.

Normally you won't run this directly. Rather, it will be launched by an
execution engine. However, it can be useful to run this on its own for testing
and debugging purposes.

options:
  --keep-stdout       Do not squelch stdout

    """
    config = load_config(args.get('<config-file>'))

    if config.get('local-imports', True):
        sys.path.insert(0, '')

    operator = cosmic_ray.plugins.get_operator(args['<operator>'])
    test_runner = cosmic_ray.plugins.get_test_runner(
        config['test-runner']['name'],
        config['test-runner']['args'])

    with open(os.devnull, 'w') as devnull:
        with redirect_stdout(sys.stdout if args['--keep-stdout'] else devnull):
            work_record = cosmic_ray.worker.worker(
                args['<module>'],
                operator,
                int(args['<occurrence>']),
                test_runner)

    sys.stdout.write(json.dumps(work_record))


DOC_TEMPLATE = """{program}

Usage: {program} [options] <command> [<args> ...]

Options:
  -h --help     Show this screen.
  -v --verbose  Use verbose logging

Available commands:
  {available_commands}

See '{program} help <command>' for help on specific commands.
"""


def common_option_handler(config):
    """Add verbose mode.

    :param config: holds the configuration values
    """
    if config['--verbose']:
        logging.basicConfig(level=logging.INFO)


def main(argv=None):
    """ Invoke the cosmic ray evaluation.

    :param argv: the command line arguments
    """
    signal.signal(signal.SIGINT,
                  lambda *args: sys.exit(0))

    try:
        dsc.main(
            'cosmic-ray',
            'cosmic-ray v.2',
            argv=argv,
            doc_template=DOC_TEMPLATE,
            common_option_handler=common_option_handler)
    except docopt.DocoptExit as exc:
        exc.code = os.EX_USAGE
        raise


if __name__ == '__main__':
    main()



"""
Facilities for counting how many mutants will be created in a
cross-product of operators and modules.
"""

from .parsing import get_ast
from .plugins import get_operator


class _CountingCore:
    """
    An operator core which simply counts how many times an operator finds a
    mutation site in a module.
    """

    def __init__(self):
        self.count = 0

    def visit_mutation_site(self, node, _, count):
        self.count += count
        return node

    @staticmethod
    def repr_args():
        return []


def _count(module_ast, op_name):
    """Count mutants for a single module-operator pair."""
    core = _CountingCore()
    operator = get_operator(op_name)(core)
    operator.visit(module_ast)
    return core.count


def count_mutants(modules, operators):
    """Count how many mutations each operator will peform on each module.

    Args:
        modules: A sequence of module objects
        operators: A sequence of operator plugin names (not operator instances)

    Returns: A dict of the form `{ module-object: {operator-name: count} }`,
        giving a per-operator count for each module.
    """
    return {
        mod: dict(
            filter(
                lambda t: t[1] > 0,
                ((op, _count(mod_ast, op))
                 for op in operators)))
        for (mod, mod_ast)
        in ((m, get_ast(m))
            for m in modules)
    }



import abc
import sys
import traceback

from cosmic_ray.work_record import WorkRecord


class TestOutcome:  # pylint: disable=too-few-public-methods
    """A enum of the possible outcomes for any mutant test run.
    """
    SURVIVED = 'survived'
    KILLED = 'killed'
    INCOMPETENT = 'incompetent'


# pylint: disable=too-few-public-methods
class TestRunner(metaclass=abc.ABCMeta):
    """Specifies the interface for test runners in the system.

    There are many ways to run unit tests in Python, and each method
    supported by Cosmic Ray should be provided by a TestRunner
    implementation.
    """

    def __init__(self, test_args):
        self._test_args = test_args

    @property
    def test_args(self):
        """The arguments for the test runner.

        This is typically just a string, but it's whatever was passed to the
        `TestRunner` initializer.
        """
        return self._test_args

    @abc.abstractmethod
    def _run(self):
        """Run all of the tests and return the results.

        The results are returned as a (success, result)
        tuple. `success` is a boolean indicating if the tests
        passed. `result` is any object that is appropriate to provide
        more information about the success/failure of the tests.
        """
        pass

    def __call__(self):
        """Call `_run()` and return a `WorkRecord` with the results.

        Returns: A `WorkRecord` with the `test_outcome` and `data` fields
            filled in.
        """
        try:
            test_result = self._run()
            if test_result[0]:
                return WorkRecord(
                    test_outcome=TestOutcome.SURVIVED,
                    data=test_result[1])
            return WorkRecord(
                test_outcome=TestOutcome.KILLED,
                data=test_result[1])
        except Exception:  # pylint: disable=broad-except
            return WorkRecord(
                test_outcome=TestOutcome.INCOMPETENT,
                data=traceback.format_exception(*sys.exc_info()))



# Generate a new configuration string.

import qprompt

from cosmic_ray.plugins import (execution_engine_names,
                                test_runner_names)

TEMPLATE = '''module: {module}

baseline: 10

exclude-modules:

test-runner:
  name: {test_runner}
  args: {test_args}

execution-engine:
  name: {engine}
'''


def new_config():
    """Prompt user for config variables and generate new config.

    Returns: A new configuration as a single string.
    """
    conf = {'module': qprompt.ask_str("Top-level module")}

    menu = qprompt.Menu()
    test_runners = test_runner_names()
    for at_pos, test_runner in enumerate(test_runners):
        menu.add(str(at_pos), test_runner)
    conf['test_runner'] = menu.show(header="Test runner",
                                    returns="desc")

    conf['test_args'] = qprompt.ask_str('Test args')

    menu = qprompt.Menu()
    for at_pos, engine_name in enumerate(execution_engine_names()):
        menu.add(str(at_pos), engine_name)
    conf['engine'] = menu.show(header="Execution engine",
                               returns="desc")
    return TEMPLATE.format(**conf)



import ast

from .operator import Operator


class ZeroIterationLoop(Operator):
    """An operator that modified for-loops to have zero iterations."""

    def visit_For(self, node):  # noqa # pylint: disable=invalid-name
        return self.visit_mutation_site(node)

    def mutate(self, node, _):
        """Modify the For loop to evaluate to None"""
        empty_list_node = ast.List(elts=[], ctx=ast.Load())
        new_node = ast.For(node.target, empty_list_node, node.body,
                           node.orelse)
        return new_node



try:
    from contextlib import redirect_stdout
except ImportError:

    import sys


    # noqa # redirect_stdout was introduced in Python 3.4
    class _RedirectStream:  # pylint: disable=too-few-public-methods
        """
            Copied from Python 3.5's implementation. See:
            https://github.com/python/cpython/commit/83935e76e35cf8d2fb9fe2599420f8adf421b884#diff-edbcdd20abc32f8b018deb2353ae925a
        """
        _stream = None

        def __init__(self, new_target):
            self._new_target = new_target
            # We use a list of old targets to make this CM re-entrant
            self._old_targets = []

        def __enter__(self):
            self._old_targets.append(getattr(sys, self._stream))
            setattr(sys, self._stream, self._new_target)
            return self._new_target

        def __exit__(self, exctype, excinst, exctb):
            setattr(sys, self._stream, self._old_targets.pop())


    # noqa # pylint: disable=invalid-name,too-few-public-methods
    class redirect_stdout(_RedirectStream):  # noqa
        """Context manager for temporarily redirecting stdout to another
        file."""
        _stream = "stdout"

try:
    # pylint: disable=unused-import, ungrouped-imports
    from contextlib import redirect_stderr
except ImportError:
    # redirect_stderr was introduced in Python 3.5
    # pylint: disable=invalid-name,too-few-public-methods
    class redirect_stderr(redirect_stdout):
        """
            Copied from Python 3.5's implementation. See:
            https://github.com/python/cpython/commit/83935e76e35cf8d2fb9fe2599420f8adf421b884#diff-edbcdd20abc32f8b018deb2353ae925a
        """

        _stream = "stderr"


def get_line_number(node):
    """Try to get the line number for `node`.

    If no line number is available, this returns "<UNKNOWN>".
    """
    if hasattr(node, 'lineno'):
        return node.lineno
    return '<UNKNOWN>'


def build_mutations(ops, to_ops):
    """The sequence of `(idx, to-op)` tuples describing the mutations for `ops`.

    Each `idx` is an index into `ops` indicating the operator to be mutated
    from. Each `to-op` is the operator class to be mutated to.

    @ops - a list of operations we want to mutate
    @to_ops - callable - yields all possible values to mutate to
    """
    # note: mutations to self are excluded.
    # None is a special mutation meaning to delete the operator!
    # 1) when to_op is None isinstance(from_op, None) will blow up because
    #    the second parameter needs to be a class
    # 2) when to_op != None we do the isinstance() check to figure out
    #    whether or not to include the operator in the list of possible
    #    mutations
    #
    # The `if to_op is None or isinstance(from_op, to_op)` expression handles
    # both scenarios very elegantly. First we handle 1) and if this is True
    # the rest of the expression is not evaluated and None is returned. Else
    # we're in scenario 2) where the left part of the expression is False so
    # the right part is evaluated. Since the left part of the expression has
    # confirmed that to_op != None then we're confident that the isinstance()
    # method will always work.
    return [(idx, to_op)
            for idx, from_op in enumerate(ops)
            for to_op in to_ops(from_op)
            if to_op is None or not isinstance(from_op, to_op)]



from cosmic_ray.work_db import use_db, WorkDB
from cosmic_ray.plugins import get_execution_engine


def execute(db_name):
    """Execute any pending work in the database stored in `db_name`, recording the
results.
    This looks for any work in `db_name` which has no results, schedules it to
    be executed, and records any results that arrive.
    """

    with use_db(db_name, mode=WorkDB.Mode.open) as work_db:
        config, timeout = work_db.get_config()
        engine_config = config['execution-engine']
        executor = get_execution_engine(engine_config['name'])
        work_records = executor(timeout,
                                work_db.pending_work,
                                config)

        for work_record in work_records:
            work_db.update_work_record(work_record)



import ast
import collections
import logging

import cosmic_ray.util

LOG = logging.getLogger()

# TODO: Can this be removed?
MutationRecord = collections.namedtuple('MutationRecord',
                                        ['module_name',
                                         'module_file',
                                         'operator',
                                         'activation_record',
                                         'mutant'])


def _full_module_name(obj):
    return '{}.{}'.format(
        obj.__class__.__module__,
        obj.__class__.__name__)


class MutatingCore:
    """
    An `Operator` core which performs mutation of ASTs.

    This core is instantiated with a target count N. The Nth time the operator
    using the core calls `visit_mutation_site()`, this core will set the
    `activation_record` attribute and perform a mutation. In other words, this
    will mutate the `target`-th instance of an operator's mutation candidates
    if such a candidate exists. If there is no `target`-th candidate then
    `activation_record` will remain `None` and no mutation will occur.
    """

    def __init__(self, target):
        self._target = target
        self._count = 0
        self._activation_record = None

    @property
    def activation_record(self):
        """The activation record for the operator.

        The activation record is a dict describing where and how the
        operator was applied.
        """
        return self._activation_record

    def visit_mutation_site(self, node, op,  # pylint: disable=invalid-name
                            num_mutations):
        """Potentially mutate `node`, returning the mutated version.

        `Operator` calls this when AST iteration reaches a
        potential mutation site. If that site is scheduled for
        mutation, the subclass instance will be asked to perform the
        mutation.
        """
        # If the current operator will do at least that many mutations,
        # then let it make the mutation now.
        if self._count <= self._target < self._count + num_mutations:
            assert self._activation_record is None
            assert self._target - self._count < num_mutations

            self._activation_record = {
                'operator': _full_module_name(op),
                'occurrence': self._target,
                'line_number': cosmic_ray.util.get_line_number(node)
            }

            old_node = node
            node = op.mutate(old_node, self._target - self._count)
            # add lineno and col_offset for newly created nodes
            ast.fix_missing_locations(node)

        self._count += num_mutations
        return node

    def repr_args(self):
        return [('target', self._target)]



import io
import os
import re
import sys

from setuptools import setup, find_packages


def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as handle:
        return handle.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version_info__ = \((.*), (.*), (.*)\)",
                              version_file, re.M)
    if version_match:
        return '.'.join(version_match.group(i) for i in range(1, 4))
    raise RuntimeError("Unable to find version string.")


LONG_DESCRIPTION = read('README.rst', mode='rt')

OPERATORS = [
    'number_replacer = '
    'cosmic_ray.operators.number_replacer:NumberReplacer',

    'mutate_comparison_operator = '
    'cosmic_ray.operators.comparison_operator_replacement:'
    'MutateComparisonOperator',

    'replace_true_false = '
    'cosmic_ray.operators.boolean_replacer:ReplaceTrueFalse',

    'replace_and_with_or = '
    'cosmic_ray.operators.boolean_replacer:ReplaceAndWithOr',

    'replace_or_with_and = '
    'cosmic_ray.operators.boolean_replacer:ReplaceOrWithAnd',

    'add_not = '
    'cosmic_ray.operators.boolean_replacer:AddNot',

    'mutate_unary_operator ='
    'cosmic_ray.operators.unary_operator_replacement:MutateUnaryOperator',

    'mutate_binary_operator ='
    'cosmic_ray.operators.binary_operator_replacement:MutateBinaryOperator',

    'break_continue_replacement ='
    'cosmic_ray.operators.break_continue:ReplaceBreakWithContinue',

    'exception_replacer ='
    'cosmic_ray.operators.exception_replacer:ExceptionReplacer',

    'zero_iteration_loop ='
    'cosmic_ray.operators.zero_iteration_loop:ZeroIterationLoop',

    'remove_decorator ='
    'cosmic_ray.operators.remove_decorator:RemoveDecorator',
]

INSTALL_REQUIRES = [
    'astunparse',
    'decorator',
    'docopt_subcommands>=2.3.0',
    'pathlib',
    'pyyaml',
    'qprompt',
    'stevedore',
    'tinydb>=3.2.1',
]

if sys.version_info < (3, 4):
    INSTALL_REQUIRES.append('enum34')

setup(
    name='cosmic_ray',
    version=find_version('cosmic_ray/version.py'),
    packages=find_packages(),

    author='Sixty North AS',
    author_email='austin@sixty-north.com',
    description='Mutation testing',
    license='MIT License',
    keywords='testing',
    url='http://github.com/sixty-north/cosmic-ray',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Testing',
    ],
    platforms='any',
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'test': ['hypothesis', 'pytest', 'tox'],
        'docs': ['sphinx', 'sphinx_rtd_theme']
    },
    entry_points={
        'console_scripts': [
            'cosmic-ray = cosmic_ray.cli:main',
            'cr-report = cosmic_ray.commands.format:report',
            'cr-rate = cosmic_ray.commands.format:format_survival_rate',
            'cr-xml = cosmic_ray.commands.format:report_xml',
        ],
        'cosmic_ray.test_runners': [
            'unittest = cosmic_ray.testing.unittest_runner:UnittestRunner',
        ],
        'cosmic_ray.operators': OPERATORS,
        'cosmic_ray.execution_engines': [
            'local = cosmic_ray.execution.local:LocalExecutionEngine',
        ]
    },
    long_description=LONG_DESCRIPTION,
)



"""This is the body of the low-level worker tool.

A worker is intended to run as a process that imports a module, mutates it in
one location with one operator, runs the tests, reports the results, and dies.
"""

import difflib
import importlib
import inspect
import json
import logging
import subprocess
import sys
import traceback

import astunparse

from .config import serialize_config
from .importing import preserve_modules, using_ast
from .mutating import MutatingCore
from .parsing import get_ast
from .testing.test_runner import TestOutcome
from .work_record import WorkRecord

LOG = logging.getLogger()


class WorkerOutcome:  # pylint: disable=too-few-public-methods
    """Possible outcomes for a worker.
    """
    NORMAL = 'normal'
    EXCEPTION = 'exception'
    NO_TEST = 'no-test'
    TIMEOUT = 'timeout'


def worker(module_name,
           operator_class,
           occurrence,
           test_runner):
    """Mutate the OCCURRENCE-th site for OPERATOR_CLASS in MODULE_NAME, run the
    tests, and report the results.

    This is fundamentally the single-mutation-and-test-run process
    implementation.

    There are three high-level ways that a worker can finish. First, it could
    fail exceptionally, meaning that some uncaught exception made its way from
    some part of the operation to terminate the function. This function will
    intercept all exceptions and return it in a non-exceptional structure.

    Second, the mutation testing machinery may determine that there is no
    OCCURENCE-th instance for OPERATOR_NAME in the module under test. In this
    case there is no way to report a test result (i.e. killed, survived, or
    incompetent) so a special value is returned indicating that no mutation is
    possible.

    Finally, and hopefully normally, the worker will find that it can run a
    test. It will do so and report back the result - killed, survived, or
    incompetent - in a structured way.

    Returns: a WorkRecord

    Raises: This will generally not raise any exceptions. Rather, exceptions
        will be reported using the 'exception' result-type in the return value.

    """
    try:
        with preserve_modules():
            module = importlib.import_module(module_name)
            module_source_file = inspect.getsourcefile(module)
            module_ast = get_ast(module)
            module_source = astunparse.unparse(module_ast)

            core = MutatingCore(occurrence)
            operator = operator_class(core)
            # note: after this step module_ast and modified_ast
            # appear to be the same
            modified_ast = operator.visit(module_ast)
            modified_source = astunparse.unparse(modified_ast)

            if not core.activation_record:
                return WorkRecord(
                    worker_outcome=WorkerOutcome.NO_TEST)

            # generate a source diff to visualize how the mutation
            # operator has changed the code
            module_diff = ["--- mutation diff ---"]
            for line in difflib.unified_diff(module_source.split('\n'),
                                             modified_source.split('\n'),
                                             fromfile="a" + module_source_file,
                                             tofile="b" + module_source_file,
                                             lineterm=""):
                module_diff.append(line)

        with using_ast(module_name, module_ast):
            rec = test_runner()

        rec.update({
            'diff': module_diff,
            'worker_outcome': WorkerOutcome.NORMAL
        })
        rec.update(core.activation_record)
        return rec

    except Exception:  # noqa # pylint: disable=broad-except
        return WorkRecord(
            data=traceback.format_exception(*sys.exc_info()),
            test_outcome=TestOutcome.INCOMPETENT,
            worker_outcome=WorkerOutcome.EXCEPTION)


def worker_process(work_record,
                   timeout,
                   config):
    """Run `cosmic-ray worker` in a subprocess and return the results,
    passing `config` to it via stdin.

    Returns: An updated WorkRecord

    """
    # The work_record param may come as just a dict (e.g. if it arrives over
    # celery), so we reconstruct a WorkRecord to make it easier to work with.
    work_record = WorkRecord(work_record)

    command = 'cosmic-ray worker {module} {operator} {occurrence}'.format(
        **work_record)

    LOG.info('executing: %s', command)

    proc = subprocess.Popen(command.split(),
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            universal_newlines=True)
    config_string = serialize_config(config)
    try:
        outs, _ = proc.communicate(input=config_string, timeout=timeout)
        result = json.loads(outs)
        work_record.update({
            k: v
            for k, v
            in result.items()
            if v is not None
        })
    except subprocess.TimeoutExpired as exc:
        work_record.worker_outcome = WorkerOutcome.TIMEOUT
        work_record.data = exc.timeout
        proc.kill()
    except json.JSONDecodeError as exc:
        work_record.worker_outcome = WorkerOutcome.EXCEPTION
        work_record.data = exc

    work_record.command_line = command
    return work_record



from .operator import Operator


class RemoveDecorator(Operator):
    """An operator that removes each of the non standard decorators."""
    REGULAR_DECORATORS = frozenset(["classmethod", "staticmethod",
                                    "abstractmethod"])

    @classmethod
    def _candidates(cls, node):
        """Get a list of indices into `node.decorator_list` of decorators of
        `node` which can be removed via mutation.
        """

        def skip(dec):
            return hasattr(dec, 'id') and dec.id in cls.REGULAR_DECORATORS

        return [i for (i, d) in enumerate(node.decorator_list)
                if not skip(d)]

    def visit_FunctionDef(self, node):  # noqa # pylint: disable=invalid-name
        decorator_candidates = self._candidates(node)

        if decorator_candidates:
            return self.visit_mutation_site(node, len(decorator_candidates))

        return node

    def mutate(self, node, idx):
        """Modify the decorator list to remove one decorator at each
         mutation"""
        candidates = self._candidates(node)
        del node.decorator_list[candidates[idx]]
        return node



import ast

from .operator import Operator
from ..util import build_mutations

# None indicates we want to delete the operator
OPERATORS = (ast.UAdd, ast.USub, ast.Invert, ast.Not, None)


def _to_ops(from_op):
    """
        The sequence of operators which `from_op` could be mutated to.
    """

    for to_op in OPERATORS:
        if to_op and isinstance(from_op, ast.Not):
            # 'not' can only be removed but not replaced with
            # '+', '-' or '~' b/c that may lead to strange results
            pass
        elif isinstance(from_op, ast.UAdd) and (to_op is None):
            # '+1' => '1' yields equivalent mutations
            pass
        else:
            yield to_op


class MutateUnaryOperator(Operator):
    """An operator that modifies unary operators."""

    def visit_UnaryOp(self, node):  # pylint: disable=invalid-name
        """
            http://greentreesnakes.readthedocs.io/en/latest/nodes.html#UnaryOp
        """
        return self.visit_mutation_site(
            node,
            len(build_mutations([node.op], _to_ops)))

    def mutate(self, node, idx):
        _, to_op = build_mutations([node.op], _to_ops)[idx]
        if to_op:
            node.op = to_op()
            return node
        return node.operand
